#include "config.h"
#include <gnome.h>
#include "global.h"

enum { CURRENT_CONFIG_VERSION = 1 };

int is_config_convert_needed(void)
{
	gint	version;

	if (! gnome_config_has_section("/gsokoban/main"))
		return 0;

	version = gnome_config_get_int("/gsokoban/main/config_version=0");
	if (version != CURRENT_CONFIG_VERSION)
		return 1;
	else
		return 0;
}

static void del_str(const char *key)
{
	char	*delstr;

	delstr = g_strconcat("/gsokoban/main/", key, NULL);
	gnome_config_clean_key(delstr);
	g_free(delstr);
}

void config_convert(void)
{
	gboolean	def;
	gint	number;
	void	*config_p;
	char	*key, *value;
	char	tmp[512];
	char	*cfgstr;

	g_print("\nGSokoban configuration converter\n");
	g_print("--------------------------------\n\n");
	g_print("Converting from version 0 to version 1.\n\n");
	
	g_print("[main] -> score_by_moves\n");
	g_print("\tChanging from int to boolean\n");
	
	number = gnome_config_get_int("/gsokoban/main/score_by_moves=0");
	gnome_config_clean_key("/gsokoban/main/score_by_moves");
	gnome_config_set_bool("/gsokoban/main/score_by_moves", number ? TRUE : FALSE);

	gnome_config_sync();
	
	config_p = gnome_config_init_iterator("/gsokoban/main");
	while (gnome_config_iterator_next(config_p, &key, &value)) {
		if (strstr(key, "current-level-")) {
			sscanf(key, "current-level-%s", tmp);
			del_str(key);
			cfgstr = g_strdup_printf("/gsokoban/%s/current_level", tmp);
			gnome_config_set_string(cfgstr, value);
			g_free(cfgstr);
			g_print("[main] -> %s\n", key);
			g_print("\tMoving to [%s] and renaming to current_level\n", tmp);
		} else if (strstr(key, "num_levels-")) {
			sscanf(key, "num_levels-%s", tmp);
			del_str(key);
			cfgstr = g_strdup_printf("/gsokoban/%s/num_levels", tmp);
			gnome_config_set_string(cfgstr, value);
			g_free(cfgstr);
			g_print("[main] -> %s\n", key);
			g_print("\tMoving to [%s] and renaming to num_levels\n", tmp);
		}
	}
	value = gnome_config_get_string_with_default("/gsokoban/main/map_set=Default", &def);
	if ((def == FALSE) && (strcmp(value, "howard") == 0)) {
		gnome_config_set_string("/gsokoban/main/map_set", "Howard");
		g_print("[main] -> map_set\n");
		g_print("\tChanging howard to Howard\n");
	}

	gnome_config_set_int("/gsokoban/main/config_version", CURRENT_CONFIG_VERSION);
	gnome_config_sync();
}
