/* event_funcs.c -- Contains all event functions of gsokoban.
 * Copyright (C) 1999, 2000 Andreas Persenius <ndap@swipnet.se>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For more details see the file COPYING.
 */
 
#include <config.h>
#include <gnome.h>
#include <gdk/gdkkeysyms.h>
#include "global.h"

extern GdkPixmap	*buff_pixmap;
extern enum gamemodes	gamemode;
extern PlayerInfo	pl;
extern GameSettings	game;
extern GString	*moves;
extern GString	*pushes;

/* Takes care of the keyboard input on the map. If do_move() indicates
 * a valid move, it logs the direction to *moves; Also calls
 * check_if_finished().
 */
gint input_func(GtkWidget *widget, GdkEventKey *event, gpointer data)
{
	int retval = 0;

	if (gamemode == playback) {
		if (event->keyval == GDK_Escape)
			gamemode = normal;
		return(1);
	}

	switch (event->keyval) {
		case GDK_Down:
		case GDK_j:
			if ((retval = do_move(DOWN_MOVE))) {
				g_string_append_c(moves, 'j');
				if (retval == 2) {
					g_string_append_c(pushes, 'p');
				} else {
					g_string_append_c(pushes, 'm');
				}
			}
			break;
		case GDK_Up:
		case GDK_k:
			if ((retval = do_move(UP_MOVE))) {
				g_string_append_c(moves, 'k');
				if (retval == 2) {
					g_string_append_c(pushes, 'p');
				} else {
					g_string_append_c(pushes, 'm');
				}
			}
			break;
		case GDK_Left:
		case GDK_h:
			if ((retval = do_move(LEFT_MOVE))) {
				g_string_append_c(moves, 'h');
				if (retval == 2) {
					g_string_append_c(pushes, 'p');
				} else {
					g_string_append_c(pushes, 'm');
				}
			}
			break;
		case GDK_Right:
		case GDK_l:
			if ((retval = do_move(RIGHT_MOVE))) {
				g_string_append_c(moves, 'l');
				if (retval == 2) {
					g_string_append_c(pushes, 'p');
				} else {
					g_string_append_c(pushes, 'm');
				}
			}
			break;
		case GDK_plus:
			if (pl.level < game.map.num_levels) {
				++pl.level;
				load_new_level();
				show_status_msg(_("Jumped to level %i."), pl.level);
			}
			break;
		case GDK_minus:
			if (pl.level > 1) {
				--pl.level;
				load_new_level();
				show_status_msg(_("Jumped to level %i."), pl.level);
			}
			break;
		case GDK_u:
			undo_move_cb(NULL, (gpointer) NULL);
			break;
		default:
			break;
	}
	check_if_finished(); 
	return(0);
}

gint expose_func(GtkWidget *widget, GdkEventExpose *event)
{
	gdk_draw_pixmap(widget->window, widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
		buff_pixmap, event->area.x, event->area.y, event->area.x,
		event->area.y, event->area.width, event->area.height);

	return(TRUE);
}

gint configure_func(GtkWidget *widget, GdkEventConfigure *event)
{
	if (!buff_pixmap) {
		/* FIXME: Shouldn't it be allocation.width and height here
		   instead of requisition??? */
		buff_pixmap = gdk_pixmap_new(widget->window, widget->requisition.width, widget->requisition.height, -1);
	}
	return(TRUE);
}
