#ifndef GLOBAL_H
#define GLOBAL_H

#define MAXCOLS 128
#define MAXROWS 128
#define MAPHEIGHT (game.map.rows * game.block_side_length)
#define MAPWIDTH  (game.map.columns * game.block_side_length)
#define DEFAULT_IMAGE_SET_NAME "gsokoban"
#define DEFAULT_MAP_SET_NAME "Default"

#define GSOKODEBUG 0

#define mapidx(row, col)	((row) * game.map.columns + (col))

enum { UP_MOVE, DOWN_MOVE, LEFT_MOVE, RIGHT_MOVE };

enum blocktypes {
	man,
	northman,
	southman,
	westman,
	eastman,
	floor,
	goal,
	object,
	treasure,
	wall,
	centerwall,
	east_twall,
	eastwall,
	horizwall,
	llcornerwall,
	lonewall,
	lrcornerwall,
	north_twall,
	northwall,
	south_twall,
	southwall,
	ulcornerwall,
	urcornerwall,
	vertiwall,
	west_twall,
	westwall,
	NUM_BLOCKTYPES
};

enum gamemodes {
	undefined,
	normal,
	playback
};

enum _toolbar_buttons {
	tb_restart,
	tb_undo,
	tb_sep1,
	tb_jump,
	tb_prefs,
	tb_sep2,
	tb_records,
	tb_last
};

enum _menu_game {
	menu_game_jump,
	menu_game_restart,
	menu_game_sep1,
	menu_game_records,
	menu_game_replay,
	menu_game_sep2,
	menu_game_exit,
	menu_game_last
};

typedef struct {
	gint x;
	gint y;
	gint level;
	gint moves;
	gint pushes;
} PlayerInfo;

typedef struct {
	gint	num_levels;
	gint	rows;
	gint	columns;
	gint	min_rows;
	gint	min_columns;
} MapInfo;

typedef struct {
	gchar	*image_set_name;
	gchar	*map_set_name;	/* minus ".zip" */
	gint	anim_speed;	/* steps per second */
	gint	block_side_length;
	gboolean	score_by_moves;	/* 0 == score by pushes */
	gboolean	auto_shrink;
	gboolean	show_toolbar;
	MapInfo	map;
} GameSettings;

/***************************
 *   FUNCTION PROTOTYPES   *
 ***************************/

/* gsoko.c */
void load_new_level(void);
void draw_map_to_buffer(void);
void draw_buffer_to_screen(void);
GSList *construct_image_sets_list(GSList *list);
GCompareFunc *image_set_sort_func(gconstpointer a, gconstpointer b);
void load_pixmaps(void);
gint do_move(gint direction);
gint undo_move(gint direction, gchar push);
void check_if_finished(void);
void show_status_msg(char *format, ...);
void show_error_msg(char *format, ...);

/* event_funcs.c */
gint input_func(GtkWidget *widget, GdkEventKey *event, gpointer data);
gint expose_func(GtkWidget *widget, GdkEventExpose *event);
gint configure_func(GtkWidget *widget, GdkEventConfigure *event);

/* menucb.c */
void preferences_cb(GtkWidget *widget, gpointer data);
void playback_level_cb(GtkWidget *widget, gpointer data);
void show_score_cb(GtkWidget *widget, gpointer data);
void goto_level_cb(GtkWidget *widget, gpointer data);
void undo_move_cb(GtkWidget *widget, gpointer data);
void restart_cb(GtkWidget *widget, gpointer data);
void about_cb(GtkWidget *widget, gpointer data);

/* stringcomp.c */
#include "stringcomp.h"

/* preferences.c */
void init_prefs_dialog(void);

/* mapinit.c */
gint init_map(gchar *mapfilename);
GSList *construct_map_sets_list(GSList *list);
gint get_num_levels(gchar *map_set);
gchar *read_file_from_zip(gchar *filepath, gchar *fileInZip);

/* convertconfig.c */
int is_config_convert_needed(void);
void config_convert(void);

#endif /* GLOBAL_H */
