/* gsoko.c -- Main file of gsokoban
 * Copyright (C) 1999, 2000 Andreas Persenius <ndap@swipnet.se>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For more details see the file COPYING.
 */
 
#include <config.h>
#include <gnome.h>
#include <errno.h>
#include <dirent.h>
#include <string.h>
#include "global.h"

static void draw_background(void);
static void init_gsokoban(void);
static gint load_pixmap(gchar *dir, gchar *name, GdkPixmap **dest, GdkBitmap **dest_mask);
static void do_window(void);
static int valid_move(int y, int x);
static void draw_pixmap(enum blocktypes which, gint x, gint y);
static void draw_pixmap_buffer(enum blocktypes  which, gint x, gint y);
static void set_transparency(enum blocktypes which, gint x, gint y);
static void unset_transparency(enum blocktypes which, gint x, gint y);
static void quit_cb(GtkWidget *widget, gpointer data);

GtkWidget	*window;
GnomeDockItem	*the_toolbar;
GtkWidget	*eventbox;
GtkWidget	*map_area;
GtkWidget	*statusbar;
GtkWidget	*toolbar_buttons[tb_last];

GdkPixmap	*buff_pixmap = NULL;
GdkPixmap	*block_pixmap[NUM_BLOCKTYPES];
GdkBitmap	*block_pixmap_mask[NUM_BLOCKTYPES];
gboolean	block_pixmap_allocated[NUM_BLOCKTYPES];

/*enum blocktypes map[game.map.rows][game.map.columns];
enum blocktypes origmap[game.map.rows][game.map.columns];*/ /* keeps track of the goals */
enum blocktypes *map, *origmap;
gint	statusbar_id;
gint	goalcount;
GString	*moves = NULL;	/* records all moves on a level */
GString	*pushes = NULL;	/* records all pushes on a level */
GSList	*map_sets;
GSList	*image_sets;

/* fallback here is an index of previously loaded image, unless fallback == -1
   if fallback => current index, special case is taken, which is:
   any image later that falls back here and is loaded, the same pointer is
   set here. block_pixmap_allocated above takes care that no images are
   extra freed  XXX */
const struct ImageInfo {
	gchar * image_name;
	int     fallback;
} image_info[NUM_BLOCKTYPES] = {
	{ "man",		-2 },
	{ "northman",		man },
	{ "southman",		man },
	{ "westman",		man },
	{ "eastman",		man },
	{ "floor", 		-1 },
	{ "goal", 		-1 },
	{ "object", 		-1 },
	{ "treasure", 		object },
	{ "wall",		-2 },
	{ "centerwall",		wall },
	{ "east_twall",		wall },
	{ "eastwall",		wall },
	{ "horizwall",		wall },
	{ "llcornerwall",	wall },
	{ "lonewall",		wall },
	{ "lrcornerwall",	wall },
	{ "north_twall",	wall },
	{ "northwall",		wall },
	{ "south_twall",	wall },
	{ "southwall",		wall },
	{ "ulcornerwall",	wall },
	{ "urcornerwall",	wall },
	{ "vertiwall",		wall },
	{ "west_twall",		wall },
	{ "westwall",		wall } };

enum gamemodes	gamemode = undefined;
PlayerInfo	pl;
GameSettings	game;

GnomeUIInfo	filemenu[] = {
	{ GNOME_APP_UI_ITEM, N_("Change level..."), N_("Jump to another level"),
		(gpointer) goto_level_cb, NULL, NULL, GNOME_APP_PIXMAP_STOCK,
		GNOME_STOCK_MENU_JUMP_TO, 0, 0, NULL },
	{ GNOME_APP_UI_ITEM, N_("Restart level"), N_("Restart the level"),
		(gpointer) restart_cb, NULL, NULL, GNOME_APP_PIXMAP_STOCK,
		GNOME_STOCK_MENU_REFRESH, 'P', GDK_CONTROL_MASK, NULL },
	GNOMEUIINFO_SEPARATOR,
	{ GNOME_APP_UI_ITEM, N_("Records..."), N_("Show level records"),
		(gpointer) show_score_cb, NULL, NULL, GNOME_APP_PIXMAP_STOCK,
		GNOME_STOCK_MENU_SCORES, 0, 0, NULL },
	{ GNOME_APP_UI_ITEM, N_("Replay solution"), N_("Replays an old level"),
		(gpointer) playback_level_cb, NULL, NULL, GNOME_APP_PIXMAP_STOCK,
		GNOME_STOCK_MENU_FORWARD, 0, 0, NULL },
	GNOMEUIINFO_SEPARATOR,
	{ GNOME_APP_UI_ITEM, N_("Exit"), N_("Exit GSokoban"),
		(gpointer) quit_cb, NULL, NULL, GNOME_APP_PIXMAP_STOCK,
		GNOME_STOCK_MENU_EXIT, 'Q', GDK_CONTROL_MASK, NULL },
	GNOMEUIINFO_END
};

GnomeUIInfo	settingsmenu[] = {
	{ GNOME_APP_UI_ITEM, N_("Preferences..."), N_("Edit preferences"),
		(gpointer) preferences_cb, NULL, NULL, GNOME_APP_PIXMAP_STOCK,
		GNOME_STOCK_MENU_PROP, 0, 0, NULL },
	GNOMEUIINFO_END
};


GnomeUIInfo	helpmenu[] = {
	{ GNOME_APP_UI_ITEM, N_("About"), N_("About GSokoban"),
		(gpointer) about_cb, NULL, NULL, GNOME_APP_PIXMAP_STOCK,
		GNOME_STOCK_MENU_ABOUT, 0, 0, NULL },
	GNOMEUIINFO_END
};

GnomeUIInfo	mainmenu[] = {
	GNOMEUIINFO_SUBTREE(N_("_Game"), filemenu),
	GNOMEUIINFO_SUBTREE(N_("_Settings"), settingsmenu),
	GNOMEUIINFO_SUBTREE(N_("H_elp"), helpmenu),
	GNOMEUIINFO_END
};

/* If changing the order of the buttons, don't forget to update
 * toolbar_buttons[] in do_window().
 **/
GnomeUIInfo	toolbar[] = {
	{ GNOME_APP_UI_ITEM, N_("Restart"), N_("Restart the level"),
		(gpointer) restart_cb, NULL, NULL, GNOME_APP_PIXMAP_STOCK,
		GNOME_STOCK_PIXMAP_REFRESH, 'R', GDK_CONTROL_MASK, NULL },
	{ GNOME_APP_UI_ITEM, N_("Undo"), N_("Undo a move"),
		(gpointer) undo_move_cb, NULL, NULL, GNOME_APP_PIXMAP_STOCK,
		GNOME_STOCK_PIXMAP_UNDO, 0, 0, NULL},
	GNOMEUIINFO_SEPARATOR,
	{ GNOME_APP_UI_ITEM, N_("Jump"), N_("Jump to another level"),
		(gpointer) goto_level_cb, NULL, NULL, GNOME_APP_PIXMAP_STOCK,
		GNOME_STOCK_PIXMAP_JUMP_TO, 0, 0, NULL },
	{ GNOME_APP_UI_ITEM, N_("Prefs"), N_("Edit preferences"),
		(gpointer) preferences_cb, NULL, NULL, GNOME_APP_PIXMAP_STOCK,
		GNOME_STOCK_PIXMAP_PREFERENCES, 0, 0, NULL },
	GNOMEUIINFO_SEPARATOR,
	{ GNOME_APP_UI_ITEM, N_("Records"), N_("Show level records"),
		(gpointer) show_score_cb, NULL, NULL, GNOME_APP_PIXMAP_STOCK,
		GNOME_STOCK_PIXMAP_SCORES, 0, 0, NULL },
	GNOMEUIINFO_END
};

int main(int argc, char *argv[])
{
	gnome_init("gsokoban", VERSION, argc, argv);
	
	init_gsokoban();
	load_pixmaps();
	do_window();
	
	gtk_widget_show(window);
	
	if (game.show_toolbar == TRUE)
		gtk_widget_show(GTK_WIDGET(the_toolbar));
	else
		gtk_widget_hide(GTK_WIDGET(the_toolbar));
	
	load_new_level();

	gtk_main();
	
	return(0);
}

void load_new_level(void)
{
	gint	retval;
	gchar	*cfgstr;

	pl.moves = 0;
	pl.pushes = 0;
	goalcount = 0;
	g_string_truncate(moves, 0);
	g_string_truncate(pushes, 0);
	cfgstr = g_strdup_printf("/gsokoban/%s/current_level",
	                         game.map_set_name);
	gnome_config_set_int(cfgstr, pl.level);
	retval = init_map(NULL);
	if (retval == 2) {	/* map could no be loaded */
		return;
	} else if (retval == 1) {	/* need to enlarge map */
		gdk_pixmap_unref(buff_pixmap);
		if (GSOKODEBUG) g_print("Creating new pixmap, size %ix%i (block size %i)\n", MAPWIDTH, MAPHEIGHT, game.block_side_length);
		buff_pixmap = gdk_pixmap_new(map_area->window, MAPWIDTH,
			MAPHEIGHT, -1);
		gtk_drawing_area_size(GTK_DRAWING_AREA(map_area), MAPWIDTH, MAPHEIGHT);
		init_map(NULL);
	} 
	draw_background();
	draw_map_to_buffer();
	draw_buffer_to_screen();
	gamemode = normal;
	g_free(cfgstr);
}

void draw_map_to_buffer(void)
{
	guint	i, j;

	for (i = 0; i < game.map.rows; i++)
		for (j = 0; j < game.map.columns; j++) {
			draw_pixmap_buffer(floor, j, i);
			if (origmap[mapidx(i, j)] == goal) {
				draw_pixmap_buffer(goal, j, i);
				if (map[mapidx(i, j)] == object)
					draw_pixmap_buffer(treasure, j, i);
			} else if (map[mapidx(i, j)] != floor)  {
				draw_pixmap_buffer(map[mapidx(i, j)], j, i);
			}
		}
	draw_pixmap_buffer(man, pl.x, pl.y);
}

static void draw_background(void)
{
	gint	i, j;

	for (i = 0; i < game.map.rows; i++)
		for (j = 0; j < game.map.columns; j++)
			gdk_draw_pixmap(buff_pixmap, map_area->style->fg_gc[GTK_WIDGET_STATE(map_area)],
					block_pixmap[floor], 0, 0,
					j * game.block_side_length,
					i * game.block_side_length,
					game.block_side_length,
					game.block_side_length);
}

void draw_buffer_to_screen(void)
{
	gdk_draw_pixmap(map_area->window,
		map_area->style->fg_gc[GTK_WIDGET_STATE(map_area)],
		buff_pixmap, 0, 0, 0, 0, MAPWIDTH, MAPHEIGHT);
}

static void init_gsokoban(void)
{
	gchar	*cfgstr;
	
	bindtextdomain(PACKAGE, GNOMELOCALEDIR);
	textdomain(PACKAGE);
	
	gdk_imlib_init();
	
	if (is_config_convert_needed())
		config_convert();
	
	/* Name of current map set */
	cfgstr = g_strdup_printf("/gsokoban/main/map_set=%s", DEFAULT_MAP_SET_NAME);
	game.map_set_name = gnome_config_get_string(cfgstr);
	g_free(cfgstr);
	/* Current level */
	cfgstr = g_strdup_printf("/gsokoban/%s/current_level=1", game.map_set_name);
	pl.level = gnome_config_get_int(cfgstr);
	g_free(cfgstr);
	/* Create an empty moves list */
	moves = g_string_new(NULL);
	/* Create an empty pushes list */
	pushes = g_string_new(NULL);
	/* Animation speed */
	game.anim_speed = gnome_config_get_int("/gsokoban/main/anim_speed=10");
	/* Minimum dimensions of map */
	game.map.min_rows = gnome_config_get_int("/gsokoban/main/min_rows=17");
	game.map.min_columns = gnome_config_get_int("/gsokoban/main/min_columns=20");
	/* Dimension of the map */
	cfgstr = g_strdup_printf("/gsokoban/main/map_cols=%i", game.map.min_columns);
	game.map.columns = gnome_config_get_int(cfgstr);
	g_free(cfgstr);
	cfgstr = g_strdup_printf("/gsokoban/main/map_rows=%i", game.map.min_rows);
	game.map.rows = gnome_config_get_int(cfgstr);
	g_free(cfgstr);
	/* Name of the current graphic set */
	cfgstr = g_strdup_printf("/gsokoban/main/image_set_name=%s", DEFAULT_IMAGE_SET_NAME);
	game.image_set_name = gnome_config_get_string(cfgstr);
	g_free(cfgstr);

	/* Any new maps? */
	cfgstr = g_strdup_printf("/gsokoban/%s/num_levels=0", game.map_set_name);
	game.map.num_levels = gnome_config_get_int(cfgstr);
	g_free(cfgstr);
	game.map.num_levels = get_num_levels(game.map_set_name);
	cfgstr = g_strdup_printf("/gsokoban/%s/num_levels", game.map_set_name);
	gnome_config_set_int(cfgstr, game.map.num_levels);
	g_free(cfgstr);
	gnome_config_sync();

	game.score_by_moves = gnome_config_get_bool("/gsokoban/main/score_by_moves=FALSE");
	game.auto_shrink = gnome_config_get_bool("/gsokoban/main/auto_shrink=FALSE");
	game.show_toolbar = gnome_config_get_bool("/gsokoban/main/toolbar=TRUE");
}

static void do_window(void)
{
	GtkWidget	*hbox;
	GtkWidget	*vbox;
		
	window = gnome_app_new("gsokoban", "GSokoban");
	gtk_widget_realize(window);
	gtk_window_set_policy(GTK_WINDOW(window), FALSE, TRUE, TRUE);
	gtk_signal_connect(GTK_OBJECT(window), "delete_event", GTK_SIGNAL_FUNC(quit_cb), NULL);
	gtk_signal_connect(GTK_OBJECT(window), "key_press_event", GTK_SIGNAL_FUNC(input_func), NULL);

	gnome_app_create_menus(GNOME_APP(window), mainmenu);
	
	gnome_app_create_toolbar(GNOME_APP(window), toolbar);
	toolbar_buttons[tb_restart] = toolbar[0].widget;
	toolbar_buttons[tb_undo] = toolbar[1].widget;
	toolbar_buttons[tb_jump] = toolbar[3].widget;

	the_toolbar = gnome_app_get_dock_item_by_name(GNOME_APP(window),
			GNOME_APP_TOOLBAR_NAME);
	
	hbox = gtk_hbox_new(TRUE, GNOME_PAD);
	gnome_app_set_contents(GNOME_APP(window), hbox);
	
	gtk_widget_push_visual(gdk_imlib_get_visual());
	gtk_widget_push_colormap(gdk_imlib_get_colormap());
	map_area = gtk_drawing_area_new();
	gtk_widget_pop_colormap();
	gtk_widget_pop_visual();
	
	vbox = gtk_vbox_new(TRUE, GNOME_PAD);
	gtk_box_pack_start(GTK_BOX(hbox), vbox, TRUE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), map_area, TRUE, FALSE, 0);
	gtk_drawing_area_size(GTK_DRAWING_AREA(map_area), MAPWIDTH, MAPHEIGHT);
	gtk_signal_connect(GTK_OBJECT(map_area), "expose_event", GTK_SIGNAL_FUNC(expose_func), NULL);
	gtk_signal_connect(GTK_OBJECT(map_area), "configure_event", GTK_SIGNAL_FUNC(configure_func), NULL);
	gtk_widget_set_events(map_area, gtk_widget_get_events(map_area) | (GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK));

	statusbar = gtk_statusbar_new();
	statusbar_id = gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar), "gsokoban");
	gnome_app_set_statusbar(GNOME_APP(window), statusbar);
	show_status_msg(_("Welcome to GSokoban!"));

	gnome_app_install_menu_hints(GNOME_APP(window), mainmenu);
	
	gtk_widget_show_all(window);
}

static void draw_pixmap(enum blocktypes which, gint x, gint y)
{
	set_transparency(which, x, y);
	gdk_draw_pixmap(map_area->window, map_area->style->fg_gc[GTK_WIDGET_STATE(map_area)], block_pixmap[which], 0, 0, x * game.block_side_length, y * game.block_side_length, game.block_side_length, game.block_side_length);
	gdk_draw_pixmap(buff_pixmap, map_area->style->fg_gc[GTK_WIDGET_STATE(map_area)], block_pixmap[which], 0, 0, x * game.block_side_length, y * game.block_side_length, game.block_side_length, game.block_side_length);
	unset_transparency(which, x, y);
}

/* 
 * Only used when initializing the buffer pixmap
 */
static void draw_pixmap_buffer(enum blocktypes which, gint x, gint y)
{
	set_transparency(which, x, y);
	gdk_draw_pixmap(buff_pixmap, map_area->style->fg_gc[GTK_WIDGET_STATE(map_area)], block_pixmap[which], 0, 0, x * game.block_side_length, y * game.block_side_length, game.block_side_length, game.block_side_length);
	unset_transparency(which, x, y);
}

static void set_transparency(enum blocktypes which, gint x, gint y)
{
	gdk_gc_set_clip_origin(map_area->style->fg_gc[GTK_WIDGET_STATE(map_area)],
		x * game.block_side_length, y * game.block_side_length);
	gdk_gc_set_clip_mask(map_area->style->fg_gc[GTK_WIDGET_STATE(map_area)],
		block_pixmap_mask[which]);
}

static void unset_transparency(enum blocktypes which, gint x, gint y)
{
	gdk_gc_set_clip_origin(map_area->style->fg_gc[GTK_WIDGET_STATE(map_area)],
		0, 0);
	gdk_gc_set_clip_mask(map_area->style->fg_gc[GTK_WIDGET_STATE(map_area)],
		NULL);
}

/**
 * construct_image_sets_list:
 * @list: a (usually) empty list to populate
 *
 * Searches the GSokoban pixmap directory for subdirs containing map sets.
 *
 * Returns a GSList with the names of all the image sets available to
 * GSokoban. The list is sorted alphabetically.
 */
GSList *construct_image_sets_list(GSList *list)
{
	gchar	*dirname, *subdir;
	DIR	*dir;
	char	*entry;
	struct dirent	*dentry;

	dirname = gnome_unconditional_pixmap_file("gsokoban");
	dir = opendir(dirname);

	while ((dentry = readdir(dir)) != NULL) {
		entry = g_strdup(dentry->d_name);
		if (!strcmp(entry, ".") || !strcmp(entry, "..")) {
			g_free(entry);
			continue;
		}
		subdir = g_strconcat(dirname, "/", entry, NULL);
		if (!g_file_test(subdir, G_FILE_TEST_ISDIR)) {
			g_free(subdir);
			g_free(entry);
			continue;
		} else 
			g_free(subdir);
		list = g_slist_insert_sorted(list, (gpointer) entry, (GCompareFunc) image_set_sort_func);
	}
	closedir(dir);
	g_free(dirname);
	return(list);
}

GCompareFunc *image_set_sort_func(gconstpointer a, gconstpointer b)
{
	return((GCompareFunc *) strcmp((char *) a, (char *) b));
}

void load_pixmaps(void)
{
	gchar	*s, *dirname;
	gint	i, fb;

	s = g_strconcat("gsokoban/", game.image_set_name, NULL);
	dirname = gnome_pixmap_file(s);
	g_free(s);

	if (dirname == NULL) {
		dirname = gnome_unconditional_pixmap_file(g_strconcat("gsokoban/",
					game.image_set_name, NULL));
		show_error_msg(_("The directory %s/ does not exist, so no "
				"graphics could be loaded.\n\n"
				"Make sure you have installed GSokoban "
				"correctly, or started it with:\n"
				"GNOMEDIR=/prefix/to/gsokoban gsokoban"),
				dirname);
		g_free(dirname);
		return;
	}

	game.block_side_length = 0;
	for (i = 0; i < NUM_BLOCKTYPES; ++i) {
		if (block_pixmap_allocated[i]) {
			gdk_pixmap_unref(block_pixmap[i]);
			block_pixmap[i] = NULL;
			if (block_pixmap_mask[i]) {
				gdk_bitmap_unref(block_pixmap_mask[i]);
				block_pixmap_mask[i] = NULL;
			}
			block_pixmap_allocated[i] = FALSE;
		}
		fb = image_info[i].fallback;
		if (!load_pixmap(dirname, image_info[i].image_name, &block_pixmap[i], &block_pixmap_mask[i])) {
			if (fb == -1) { /* no fallback for this image */
				show_error_msg(_("Error loading image %s: %s."),
					   image_info[i].image_name,
					   g_unix_error_string(errno));
			} else if (fb == -2) { /* basic image */
/*  				g_print("load_image(): Hoping for extended images.\n"); */
			} else if (fb > i) {
				/* images that have a fallback should come
				   after the fallback image in image_info */
				g_assert_not_reached();
			} else { /* trying fallback image */
				if (! block_pixmap[fb]) { /* fallback failed */
					show_error_msg("Error loading image %s: %s.\n"
					"Tried to load fallback image %s, but that failed too.",
					image_info[i].image_name,
					g_unix_error_string(errno),
					image_info[fb].image_name);
				} else { /* yes, fallback exists */
					block_pixmap[i] = block_pixmap[fb];
					block_pixmap_mask[i] = block_pixmap_mask[fb];
				}
			}
		} else if ((fb >= 0) && !block_pixmap[fb]) {
			block_pixmap[fb] = block_pixmap[i];
			block_pixmap_mask[fb] = block_pixmap_mask[i];
		} else {
			block_pixmap_allocated[i] = TRUE;
		}
	}
	g_free(dirname);
}

/**
 * load_pixmap:
 * @dir:
 * @name:
 * @dest:
 * @dest_mask:
 *
 * Loads pixmaps.
 *
 * Returns: 1 if successful and 0 otherwise
 */
static gint load_pixmap(gchar *dir, gchar *name, GdkPixmap **dest, GdkBitmap **dest_mask)
{
	GdkImlibImage	*image;
	GdkVisual	*visual;
	gchar	*filename;

	/* This should probably be fixed to allow other image formats than
	 * xpm. */
	filename = g_strconcat(dir,"/", name, ".xpm", NULL);
	image = gdk_imlib_load_image(filename);
	
	if (!image) {
		g_free(filename);
		return(0);
	}
	
	if (game.block_side_length && (game.block_side_length != image->rgb_height)) {
		show_error_msg(_("Pixmap has wrong size.\n(%s)\n"
				 "Size: %ix%i\nCorrect size: %ix%i"),
			filename, image->rgb_width, image->rgb_height,
			game.block_side_length, game.block_side_length);
		while (gtk_events_pending())
			gtk_main_iteration();
	} else {
		game.block_side_length = image->rgb_height;
	}
	g_free(filename);
	visual = gdk_imlib_get_visual();
	if(visual->type != GDK_VISUAL_TRUE_COLOR) {
		gdk_imlib_set_render_type(RT_PLAIN_PALETTE);
	}
	gdk_imlib_render(image, image->rgb_width, image->rgb_height);
	*dest = gdk_imlib_move_image(image);
	if (! *dest) {
		show_error_msg("Could not load %s.\n", name);
		return(0);
	}
	*dest_mask = gdk_imlib_copy_mask(image);
	gdk_imlib_destroy_image(image);
	return(1);
}

/* Nice little struct used by do_move and undo_move */
struct {
	gint	y;
	gint	x;
} movetable[4] = {
	{ -1,  0},	/* UP */
	{  1,  0},	/* DOWN */
	{  0, -1},	/* LEFT */
	{  0,  1}};	/* RIGHT */

/**
 * do_move:
 * @direction: one of {LEFT, RIGHT, UP, DOWN}_MOVE
 *
 * do_move recieves coordinates and updates the screen.
 *
 * Returns: 0 if no move was made, 1 if a normal move, 2 if we pushed a
 * object (same return value as valid_move()).
 */
gint do_move(gint direction)
{
	gint	retval=0;
	gint	x = movetable[direction].x;
	gint	y = movetable[direction].y;
	static gint	last_direction = 0;
	
	
	if ((retval = valid_move(y, x))) {
		if (retval == 2) {	/* object move */
			++pl.pushes;
			if (origmap[mapidx(pl.y+y, pl.x+x)] == goal) {
				map[mapidx(pl.y+y, pl.x+x)] = goal;
				draw_pixmap(goal, pl.x+x, pl.y+y);
				if (map[mapidx(pl.y+2*y, pl.x+2*x)] != goal)
					++goalcount;
			} else {
				map[mapidx(pl.y+y, pl.x+x)] = floor;
				draw_pixmap(floor, pl.x+x, pl.y+y);
				if (map[mapidx(pl.y+2*y, pl.x+2*x)] == goal)
					--goalcount;
			}
			map[mapidx(pl.y+2*y, pl.x+2*x)] = object;
			if (origmap[mapidx(pl.y+2*y, pl.x+2*x)] == goal)
				draw_pixmap(treasure, pl.x+2*x, pl.y+2*y);
			else
				draw_pixmap(object, pl.x+2*x, pl.y+2*y);
		}
		draw_pixmap(map[mapidx(pl.y, pl.x)], pl.x, pl.y);
		pl.x += x;
		pl.y += y;
		draw_pixmap(man + direction + 1, pl.x, pl.y);
		++pl.moves;
		show_status_msg(_("[Level %i]   Moves: %i   Pushes: %i   Blocks left: %i"),
			pl.level, pl.moves, pl.pushes, goalcount);
	} else if (direction != last_direction) {
		draw_pixmap(origmap[mapidx(pl.y, pl.x)], pl.x, pl.y);
		draw_pixmap(man + direction + 1, pl.x, pl.y);
	}
	last_direction = direction;
	return(retval);
}

/* undo_move receives moves / pushes and updates the screen.
 *    int x, int y: same meaning as in do_move
 *    Beware: the coordinates passed are the good ones to make an undo move! 
 */
int undo_move(gint direction, gchar push)
{
	gint	x = movetable[direction].x;
	gint	y = movetable[direction].y;
	
	if (push == 'p') {
		--pl.pushes;
		if (origmap[mapidx(pl.y-y, pl.x-x)] == goal) {
			map[mapidx(pl.y-y, pl.x-x)] = goal;
			draw_pixmap(goal, pl.x-x, pl.y-y);
			if (map[mapidx(pl.y, pl.x)] != goal)
				++goalcount;
		} else {
			map[mapidx(pl.y-y, pl.x-x)] = floor;
			draw_pixmap(floor, pl.x-x, pl.y-y);
			if (map[mapidx(pl.y, pl.x)] == goal)
				--goalcount;
		}
		draw_pixmap(origmap[mapidx(pl.y, pl.x)], pl.x, pl.y);
		map[mapidx(pl.y, pl.x)] = object;
	}
	if (map[mapidx(pl.y, pl.x)] == object && origmap[mapidx(pl.y, pl.x)] == goal)
		draw_pixmap(treasure, pl.x, pl.y);
	else
		draw_pixmap(map[mapidx(pl.y, pl.x)], pl.x, pl.y);
	pl.x += x;
	pl.y += y;
	draw_pixmap(man + (direction ^ 1) + 1, pl.x, pl.y);
	--pl.moves;
	show_status_msg(_("[Level %i]   Moves: %i   Pushes: %i   Blocks left: %i"),
		pl.level, pl.moves, pl.pushes, goalcount);
	return 0;
}

/* Figures out if the map is finished. If there's a record, saves it and
 * the moves. Loads the next level.
 */
void check_if_finished(void)
{
	gint	levmoves = 0, levpushes = 0;
	GString	*compressed_moves;
	GtkWidget	*msgbox;
	gchar	cfgstr[256];

	g_assert(goalcount >= 0);

	if (goalcount > 0)
		return;
	
	/* Playback mode */
	if (gamemode == playback) {
		show_status_msg(_("Level %i finished."), pl.level);
		return;
	}
	
	g_snprintf(cfgstr, 256, "/gsokoban-scores/moves-%s/level%i=0", 
		game.map_set_name, pl.level);
	levmoves = gnome_config_get_int(cfgstr);
	g_snprintf(cfgstr, 256, "/gsokoban-scores/pushes-%s/level%i=0",
		game.map_set_name, pl.level);
	levpushes = gnome_config_get_int(cfgstr);
	if (    ((pl.moves < levmoves) && game.score_by_moves) ||
		((pl.pushes < levpushes) && !game.score_by_moves) ||
		(levmoves == 0)) {
		show_status_msg(_("Level %i finished. New record: %i moves & %i pushes (old: %i moves & %i pushes)"),
			pl.level, pl.moves, pl.pushes, levmoves, levpushes);
		g_snprintf(cfgstr, 256,
			"/gsokoban-scores/moves-%s/level%i",
			game.map_set_name, pl.level);
		gnome_config_set_int(cfgstr, pl.moves);
		g_snprintf(cfgstr, 256,
			"/gsokoban-scores/pushes-%s/level%i",
			game.map_set_name, pl.level);
		gnome_config_set_int(cfgstr, pl.pushes);
		
		compressed_moves = compress_string(moves);
		g_snprintf(cfgstr, 256,
			"/gsokoban-scores/move_strings-%s/level%i",
			game.map_set_name, pl.level);
		gnome_config_set_string(cfgstr, compressed_moves->str);
		g_string_free(compressed_moves, TRUE);

		gnome_config_sync();
	}
	if (pl.level == game.map.num_levels) {
		msgbox = gnome_message_box_new(_("You have finished all levels"
			" in this map set!"), GNOME_MESSAGE_BOX_INFO,
			GNOME_STOCK_BUTTON_OK, NULL);
		gnome_dialog_set_parent(GNOME_DIALOG(msgbox),
			GTK_WINDOW(window));
		gtk_widget_show(msgbox);
	} else {
		++pl.level;
		load_new_level();
	}
}

/* returns 1 if clear, 2 if one block and 0 if impossible to move. 
 * int y and x: requested move. -1 for up or left; 1 for down or right.
 */
static int valid_move(int y, int x)
{
	gint	next = map[mapidx(pl.y+y, pl.x+x)];
	gint    secondnext = map[mapidx(pl.y+2*y, pl.x+2*x)];

	if ((next == floor) || (next == goal))
		return(1);
	else if ((next == object) && ((secondnext == floor) || (secondnext == goal)))
		return(2);
	else
		return(0);
}

/* Displays a message in the statusbar. Since we don't need a history,
 * the previous message is removed from the buffer.
 */
void show_status_msg(char *format, ...)
{
	gchar	*message;
	va_list	arg_list;

	va_start(arg_list, format);
	message = g_strdup_vprintf(format, arg_list);
	va_end(arg_list);

	gtk_statusbar_pop(GTK_STATUSBAR(statusbar), statusbar_id);
	gtk_statusbar_push(GTK_STATUSBAR(statusbar), statusbar_id, message);

	g_free(message);
}

void show_error_msg(char *format, ...)
{
	GtkWidget	*msgbox;
	gchar	*message;
	va_list	arg_list;

	va_start(arg_list, format);
	message = g_strdup_vprintf(format, arg_list);
	va_end(arg_list);
	
	msgbox = gnome_message_box_new(message, GNOME_MESSAGE_BOX_ERROR,
		GNOME_STOCK_BUTTON_OK, NULL);
	if (window) {
		gnome_dialog_set_parent(GNOME_DIALOG(msgbox),
			GTK_WINDOW(window));
	}
	gtk_widget_show(msgbox);
	g_free(message);
}

static void quit_cb(GtkWidget *widget, gpointer data)
{
	g_string_free(moves, TRUE);
	g_string_free(pushes, TRUE);
	gnome_config_sync();
	gtk_main_quit();
	return;
}
