/* stringcomp.c -- String compression functions of gsokoban
 * Copyright (C) 1999, 2000 Andreas Persenius <ndap@swipnet.se>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For more details see the file COPYING.
 */
 
#include <config.h>
#include <gnome.h>
#include <ctype.h>
#include "stringcomp.h"

/*
 * Receives an uncompressed string *string which it compresses and returns
 * The returned string must be freed with g_string_free()
 */
GString *compress_string(GString *string)
{
	int	i;
	int	count = 0;
	gchar	number[3];
	gchar	character = '\0';
	GString	*tmp;

	tmp = g_string_new(NULL);	
	
	for (i = 0; i < string->len; ++i) {
		if (count > 0) {
			if (string->str[i] == character)
				++count;
			else {
				if (count > 1) {
					g_snprintf(number, 3, "%i", count);
					g_string_append(tmp, number);
				}
				g_string_append_c(tmp, character);
				character = string->str[i];
				count = 1;
			}
		} else {
			character = string->str[i];
			++count;
		}
	}
	if (count > 1) {
		g_snprintf(number, 3, "%i", count);
		g_string_append(tmp, number);
	}
	g_string_append_c(tmp, character);
	return(tmp);
}

/*
 * Receives a compressed string *string which it uncompresses and returns
 * The returned string must be freed with g_string_free()
 */
GString *uncompress_string(GString *string)
{
	int	count;
	int	i, j;
	GString	*tmp;
	
	count = 0;
	tmp = g_string_new(NULL);

	for (i = 0; i < string->len; ++i) {
		if (isalpha(string->str[i])) {
			if (count) {
				for (j = 0; j < count; j++)
					g_string_append_c(tmp, string->str[i]);
			} else
				g_string_append_c(tmp, string->str[i]);
			j = count = 0;
		} else {
			count = atoi(&string->str[i]);
			if (count > 9)
				++i;
		}
	}
	return(tmp);
}
