#if 0			/* lines below are for testing */
NAME=`basename $0 .c`
[ ! $1 ] && set -- -g -O2
CMDLN="${CC:-gcc} -Wall $@ -o $NAME $NAME.c -DTESTMAIN -I/usr/lib/glib/include"
echo $CMDLN; exec $CMDLN; exit -1
#endif
/*
 * $Id: zip08x.c $
 *
 * Author: Tomi Ollila <too@iki.fi>
 *
 * 	Copyright (c) 1997 Tomi Ollila
 * 	    All rights reserved
 *
 * Created: Thu Jul 31 18:19:30 1997 too
 * Last modified: Wed Jun 16 17:48:45 1999 too
 *
 * HISTORY 
 * $Log: $
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdarg.h>
#include <sys/stat.h>

#include "zipx_priv.h"

/*
 * Make 32/16 bit variables from octet string, in little endian format.
 */

/* defined to makelong() in zipx_priv.h for internal usage */
guint32 __zip08x_internal_makelong(unsigned char * s)
{
  return ((guint32)s[3] << 24) | ((guint32)s[2] << 16)
    |    ((guint32)s[1] << 8)  |  (guint32)s[0];
}

/* defined to makeword() in zipx_priv.h for internal usage */
guint16 __zip08x_internal_makeword(unsigned char * s)
{
    return ((guint16)s[1] << 8) | (guint16)s[0];
}

#	define NUM_DIRECTORY_ENTRIES	8
#	define SIZE_DIRECTORY		12
#	define OFFSET_DIRECTORY_START	16

#	define METHOD_COMPRESSION	10
#	define CRC32_FILE		16
#	define SIZE_COMPRESSED		20
#	define SIZE_UNCOMPRESSED	24
#	define LENGTH_FILENAME		28
#	define LENGTH_EXTRA_FIELD	30
#	define LENGTH_FILE_COMMENT	32
#	define OFFSET_FILE		42
#	define NAME_FILE		46


#ifndef O_BINARY
#define O_BINARY 0
#endif

static long ffilelen(int fd)
{
  struct stat st;

  if (fstat(fd, &st) < 0)
    return -1;

  return st.st_size;
}

struct carrydata
{
  struct zipfileinfo * zfi;
  char * buf32k;
  short entries;
  long size;
  long offset;
};

static int find_eod(int fd, int filesize, struct carrydata * cd);
static int parse_zdir(int fd, struct carrydata * cd);

static void delete_zipfile(ZipFile * zf)
{
  if (zf->zfp)		free(zf->zfp);
  if (zf->fd >= 0)	close(zf->fd);
  if (zf->buf32k)	free(zf->buf32k);
  if (zf->zfi)		free(zf->zfi);
  free(zf);
}

static ZipFile * seterr(zrerror_t * errcode_p, ZipFile * zf, int code)
{
  if (zf)
    delete_zipfile(zf);

  if (errcode_p)
    *errcode_p = code;

  return NULL;
}

int expungeZip(ZipFile * zf)
{
  if (zf->refcount)
    return zf->refcount;

  delete_zipfile(zf);
  return 0;
}

ZipFile * openZip(char * filename, zrerror_t * errcode_p)
{
  long filesize;
  struct carrydata cdi = { 0 };
  int rv;
  ZipFile * zf;

  if ((zf = (ZipFile *)calloc(1, sizeof *zf)) == NULL)
    return seterr(errcode_p, NULL, ZRE_OUTOFMEM);

  zf->fd = -1;
  
  if ((zf->buf32k = (char *)malloc(BUF32KSIZE)) == NULL)
    return seterr(errcode_p, zf, ZRE_OUTOFMEM);    

  cdi.buf32k = zf->buf32k;
  
  if ((zf->fd = open(filename, O_RDONLY|O_BINARY)) < 0)
    return seterr(errcode_p, zf, ZRE_ZF_OPEN);

  if ((filesize = ffilelen(zf->fd)) < 0)
    return seterr(errcode_p, zf, ZRE_ZF_STAT);

  if ((rv = find_eod(zf->fd, filesize, &cdi)) != 0)
    return seterr(errcode_p, zf, rv);

#ifdef TESTMAIN
  
  printf("num_directory_entries: %d\n",	  cdi.entries);
  printf("size_directory: %ld\n",	  cdi.size);
  printf("offset_directory_start: %ld \n", cdi.offset);

#endif

  if ((rv = parse_zdir(zf->fd, &cdi)) != 0)
    return seterr(errcode_p, zf, rv);

  zf->zfi = cdi.zfi;
  zf->zdp = cdi.zfi;
  
  return zf;
}


#define ECDREADSIZE 1024
/* #define ECDREADSIZE 25 /* for testing */
/* #define BUF32KSIZE  32768 defined in zipx_priv.h */

static int find_eod(int fd, int filesize, struct carrydata * cd)
{
  long offset;
  long bufptr = BUF32KSIZE;
  int offoff;
  int firstoff = TRUE;
  int tmp;
  
  if (filesize < 22)
    return ZRE_ZF_TOO_SHORT;

  for (offset = filesize; filesize; filesize = offset)
  {
    int scanoff;
    char * buf;
    
    offoff = filesize < ECDREADSIZE? filesize: ECDREADSIZE;

    offset -= offoff;
    bufptr -= offoff;

#if 0    
    printf("%d %d %d\n", offoff, filesize, offset);
#endif    
    
    if (bufptr < 0)
      break;

    if (lseek(fd, offset, SEEK_SET) < 0)
      return ZRE_ZF_SEEK;

    buf = cd->buf32k + bufptr;
    
    if ((tmp = read(fd, buf, offoff)) < offoff)
      return ZRE_ZF_READ;

    /* ecd header is at least 22 bytes back from end of file...
       making sure buf[scanoff + x] does not scan over allocated memory */
    if (firstoff)
    {
      offoff -= 20;
      firstoff = FALSE;
    }

    for (scanoff = offoff - 1; scanoff >= 0; scanoff--)
      if (buf[scanoff] == 'P' && buf[scanoff + 1] == 'K' && 
	  buf[scanoff + 2] == '\005' &&
	  buf[scanoff + 3] == '\006')
      {
	buf += scanoff;

	cd->entries = makeword(&buf[NUM_DIRECTORY_ENTRIES]);
	cd->size = makelong(&buf[SIZE_DIRECTORY]);
	cd->offset = makelong(&buf[OFFSET_DIRECTORY_START]);
	
	return 0;
      }
  }
  return ZRE_EDH_MISSING;
}

/*
 * making pointer alignments to values that can be handled as structures
 * is tricky. First to know how much an compiler+architecture needs for
 * structure alignment, a simple spesific structure is created.
 * Since binary operations cannot be done for pointer variables,
 * GLIB convenience macros are used to convert those to ints (and back)
 * in a portable way.
 */

static int advance_pointer_keep_struct_aligned(void ** p, int i)
{
  struct { long x;  } at[2];
  int  av = GPOINTER_TO_INT(&at[1]) - GPOINTER_TO_INT(&at[0]);
  gint ip = GPOINTER_TO_INT(*p);
  gint sp = ip;

  ip = ( ip + i + (av - 1) ) & ~(av - 1);

  *p = GINT_TO_POINTER(ip);
  return ip - sp;
}

#if defined(TESTMAIN)
static void xbuf(char * buf)
{
  int i;

  for (i = 0; i < 50; i++)
  {
    int j = (unsigned char)buf[i];
    char x = isprint(j)? j: '.';
    printf("%2d: %02x %c        ", i, j, x);
  }
}
#endif

static int parse_zdir(int fd, struct carrydata * cd)
{
  struct zipfileinfo * zfi, * firstzfi;
  guint16 * lastnextp = NULL;
  short entries/*, realentries = 0*/;
  long offset;
  char * buf32k = cd->buf32k;
  
  if (cd->size > BUF32KSIZE)
    return ZRE_DIRSIZE;
  
  if (lseek(fd, cd->offset, SEEK_SET) < 0)
    return ZRE_ZF_SEEK;

  if (read(fd, buf32k, cd->size) < cd->size)
    return ZRE_ZF_READ;


  advance_pointer_keep_struct_aligned(
    (zfi = (struct zipfileinfo *)buf32k, (void**)&zfi), 0);
  
  firstzfi = zfi;
  
  for (entries=cd->entries, offset = 0; entries > 0; entries--)
  {
    char * buf = buf32k + offset;

    guint16 efsize = makeword(&buf[LENGTH_EXTRA_FIELD]);
    guint16 fcsize = makeword(&buf[LENGTH_FILE_COMMENT]);
    guint16 fnsize = makeword(&buf[LENGTH_FILENAME]);
    guint16  compr = makeword(&buf[METHOD_COMPRESSION]);

    if (compr > 255)
      compr = 255;

#if defined(TESTMAIN) && 0
    xbuf(buf);
#endif    
#if 0
    printf("offset %ld, size %ld, buf %p, zfi %p\n",
	   offset, cd->size, buf, zfi);
#endif
    
    offset+= 46 + efsize + fcsize + fnsize;

    if (offset > cd->size)
      break;

    /* writes over the read buffer, Since the structure where data is
       copied is smaller than the data in buffer this can be done.
       It is important that the order of setting the fields is considered
       when filling the structure, so that some data is not trashed in
       first structure read.
       at the end the whole copied list of structures  is copied into
       newly allocated buffer */

    /* realentries++; */
    zfi->crc32 = makelong(&buf[CRC32_FILE]);
    zfi->csize = makelong(&buf[SIZE_COMPRESSED]);
    zfi->usize = makelong(&buf[SIZE_UNCOMPRESSED]);
    zfi->offset = makelong(&buf[OFFSET_FILE]);
    zfi->compr = compr;
    memcpy(zfi->fname, &buf[NAME_FILE], fnsize);
    zfi->fname[fnsize] = '\0';
    
#ifdef TESTMAIN
    printf("\ncompression method: %d", compr);
    if (compr == 0) printf(" (stored)");
    else if (compr == 8) printf(" (deflated)");
    else printf(" (unknown)");
    printf("\ncrc32: %x\n", zfi->crc32);
    printf("compressed size: %d\n", zfi->csize);
    printf("uncompressed size: %d\n", zfi->usize);
    printf("filename length: %d\n", fnsize);
    printf("extra field length: %d\n", efsize);
    printf("file comment length: %d\n", fcsize);
    printf("offset of file in archive: %d\n", zfi->offset);
    printf("filename: %s\n\n", zfi->fname);

#endif
  
    lastnextp = &zfi->next;

    *lastnextp = advance_pointer_keep_struct_aligned((void **)&zfi,
						     sizeof *zfi + fnsize + 1);
  }
  if (!lastnextp)
    return 0; /* 0 (sane) entries in zip directory... */

  *lastnextp = 0; /* mark end of list */

  {
    int dsize = GPOINTER_TO_INT(zfi) - GPOINTER_TO_INT(firstzfi);

    cd->zfi = (struct zipfileinfo *)malloc(dsize);

    if (!cd->zfi)
      return ZRE_OUTOFMEM;

    memcpy(cd->zfi, firstzfi, dsize);
  }
    
  return 0;
}
  
#ifdef TESTMAIN

#if 1
void scan_zipfile(ZipFile * zf)
{
  struct zipfileinfo * zfi = zf->zfi;

  if (zf->zfi == NULL) return;
    
  while (1)
  {
    printf("\ncompression method: %d", zfi->compr);
    if (zfi->compr == 0) printf(" (stored)");
    else if (zfi->compr == 8) printf(" (deflated)");
    else printf(" (unknown)");
    printf("\ncrc32: %x\n", zfi->crc32);
    printf("compressed size: %d\n", zfi->csize);
    printf("uncompressed size: %d\n", zfi->usize);
    printf("offset of file in archive: %d\n", zfi->offset);
    printf("filename: %s\n\n", zfi->fname);

    if (zfi->next == 0)
      return;
    (char *)zfi += zfi->next;
    sleep(1);
 }
}
#endif  

int main(int argc, char ** argv)
{
  ZipFile * zf;
  char * name = "empty.zip";
  zrerror_t rv;
  
  if (argv[1] != NULL)
    name = argv[1];

  if ((zf = openZip(name, &rv)) == NULL)
    printf("error %d\n", rv);
  else
  {
    printf("----------------\n");
    scan_zipfile(zf);
  }

  return 0;
}

#endif
