/*
 * $Id: zipx.h $
 *
 * Author: Tomi Ollila <Tomi.Ollila@iki.fi>
 *
 * 	Copyright (c) 1999 Tomi Ollila
 * 	    All rights reserved
 *
 * Created: Wed Jun  2 17:33:34 1999 too
 * Last modified: Fri Jun 18 00:27:21 1999 too
 *
 * HISTORY 
 * $Log: $
 */

#ifndef ZIPX_H
#define ZIPX_H

struct zipfileinfo;

typedef enum {
/*  0 */ ZRE_NO_ERROR,	/* no error, may be used if user sets it. */
/*  1 */ ZRE_OUTOFMEM,	/* out of memory */
/*  2 */ ZRE_ZF_OPEN,	/* failed to open zipfile, see errno for details */
/*  3 */ ZRE_ZF_STAT,   /* failed to fstat zipfile, see errno for details */
/*  4 */ ZRE_ZF_SEEK,   /* failed to lseek zipfile, see errno for details */
/*  5 */ ZRE_ZF_READ,   /* failed to read zipfile, see errno for details */
/*  6 */ ZRE_ZF_TOO_SHORT,
/*  7 */ ZRE_EDH_MISSING,
/*  8 */ ZRE_DIRSIZE,
/*  9 */ ZRE_ENOENT,
/* 10 */ ZRE_UNSUPP_COMPR,
/* 11 */ ZRE_INFLATE,
/* 12 */ ZRE_CORRUPTED,
/* 13 */ ZRE_UNDEF
} zrerror_t;

/*
 * zip_open flags.
 */
#define ZOF_CASEINSENSITIVE	0x01
#define ZOF_IGNOREPATH		0x02

typedef struct ZipFile_s ZipFile;
typedef struct ZipFp_s ZipFp;

typedef struct ZipDirent_s
{
  int	 compr;	/* compression method (for user's information)*/
  int	 size;	/* file size */
  char * name;  /* file name */
} ZipDirent;

typedef struct ZipDirent_s ZipStat;

/*
 * Opening/closing zipfile for use. (zip08x.c)
 */
ZipFile *  openZip(char * filename, zrerror_t * errcode_p);
int	expungeZip(ZipFile * zf);

/*
 * Getting error strings (zip08errs.c)
 */
char * zip_errstr1(int errcode); /* error in openZip() */
char * zip_errstr(ZipFile * zf); /* error in other functions */


/*
 * Functions to grab information from ZipFile structure (if ever needed)
 */
int zipfile_errcode(ZipFile * zf);
void zipfile_seterrcode(ZipFile * zf, int errcode);
int zipfile_fd(ZipFile * zf);


/*
 * Scanning files in zip archive (zip08dir.c)
 */
int	readZipDirent(ZipFile * zf, ZipDirent * zde);
void	resetZipDir(ZipFile * zf);

/*
 * 'opening', 'closing' and reading invidual files in zip archive. (zip08file.c
 */

ZipFp * zip_open(ZipFile * zf, char * name, int flags);
void	zip_close(ZipFp * zfp);
int	zip_read(ZipFp * zfp, char * buf, int len);


/*
 * Functions to grab information from ZipFP structure (if ever needed)
 */
ZipFile * zipfp_zipfile(ZipFp * zfp);


/*
 * reading info of a single file (zip08stat.c)
 */

int	zip_stat(ZipFile * zf, char * name, ZipStat * zs, int flags);


#endif /* ZIPX_H */
